"""
   Copyright 2022 Dustin Roeder (dmroeder@gmail.com)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import socket
from struct import pack, unpack_from


class Device(object):

    def __init__(self):
        # structure of a logix device
        self.Length = None
        self.EncapsulationVersion = None
        self.IPAddress = None
        self.VendorID = None
        self.Vendor = None
        self.DeviceID = None
        self.DeviceType = None
        self.ProductCode = None
        self.Revision = None
        self.Status = None
        self.SerialNumber = None
        self.ProductNameLength = None
        self.ProductName = None
        self.State = None

    def __repr__(self):

        props = ''
        props += 'Length={}, '.format(self.Length)
        props += 'EncapsulationVersion={}, '.format(self.EncapsulationVersion)
        props += 'IPAddress={}, '.format(self.IPAddress)
        props += 'VendorID={}, '.format(self.VendorID)
        props += 'Vendor={}, '.format(self.Vendor)
        props += 'DeviceID={}, '.format(self.DeviceID)
        props += 'DeviceType={}, '.format(self.DeviceType)
        props += 'ProductCode={}, '.format(self.ProductCode)
        props += 'Revision={}, '.format(self.Revision)
        props += 'Status={}, '.format(self.Status)
        props += 'SerialNumber={}, '.format(self.SerialNumber)
        props += 'ProductNameLength={}, '.format(self.ProductNameLength)
        props += 'ProductName={}, '.format(self.ProductName)
        props += 'State={}'.format(self.State)

        return 'LGXDevice({})'.format(props)

    def __str__(self):

        ret = "{} {} {} {} {} {} {} {} {} {} {} {} {} {}".format(
                 self.Length,
                 self.EncapsulationVersion,
                 self.IPAddress,
                 self.VendorID,
                 self.Vendor,
                 self.DeviceID,
                 self.DeviceType,
                 self.ProductCode,
                 self.Revision,
                 self.Status,
                 self.SerialNumber,
                 self.ProductNameLength,
                 self.ProductName,
                 self.State)

        return ret

    @staticmethod
    def get_device(device_id):
        if device_id in devices.keys():
            return devices[device_id]
        else:
            return "Unknown"

    @staticmethod
    def get_vendor(vendor_id):
        if vendor_id in vendors.keys():
            return vendors[vendor_id]
        else:
            return "Unknown"

    @staticmethod
    def parse(data, ip_address=None):
        # we're going to take the packet and parse all
        #  the data that is in it.

        resp = Device()
        resp.Length = unpack_from('<H', data, 28)[0]
        resp.EncapsulationVersion = unpack_from('<H', data, 30)[0]

        long_ip = unpack_from('<I', data, 36)[0]
        if ip_address:
            resp.IPAddress = ip_address
        else:
            resp.IPAddress = socket.inet_ntoa(pack('<L', long_ip))

        resp.VendorID = unpack_from('<H', data, 48)[0]
        resp.Vendor = Device.get_vendor(resp.VendorID)

        resp.DeviceID = unpack_from('<H', data, 50)[0]
        resp.DeviceType = Device.get_device(resp.DeviceID)

        resp.ProductCode = unpack_from('<H', data, 52)[0]
        major = unpack_from('<B', data, 54)[0]
        minor = unpack_from('<B', data, 55)[0]
        resp.Revision = str(major) + '.' + str(minor)

        resp.Status = unpack_from('<H', data, 56)[0]
        resp.SerialNumber = hex(unpack_from('<I', data, 58)[0])
        resp.ProductNameLength = unpack_from('<B', data, 62)[0]
        resp.ProductName = str(data[63:63+resp.ProductNameLength].decode('utf-8'))

        state = data[-1:]
        resp.State = unpack_from('<B', state, 0)[0]

        return resp


# List originally came from Wireshark /epan/dissectors/packet-cip.c
devices = {0x00: 'Generic Device (deprecated)',
           0x02: 'AC Drive',
           0x03: 'Motor Overload',
           0x04: 'Limit Switch',
           0x05: 'Inductive Proximity Switch',
           0x06: 'Photoelectric Sensor',
           0x07: 'General Purpose Discrete I/O',
           0x09: 'Resolver',
           0x0C: 'Communications Adapter',
           0x0E: 'Programmable Logic Controller',
           0x10: 'Position Controller',
           0x13: 'DC Drive',
           0x15: 'Contactor',
           0x16: 'Motor Starter',
           0x17: 'Soft Start',
           0x18: 'Human-Machine Interface',
           0x1A: 'Mass Flow Controller',
           0x1B: 'Pneumatic Valve',
           0x1C: 'Vacuum Pressure Gauge',
           0x1D: 'Process Control Value',
           0x1E: 'Residual Gas Analyzer',
           0x1F: 'DC Power Generator',
           0x20: 'RF Power Generator',
           0x21: 'Turbomolecular Vacuum Pump',
           0x22: 'Encoder',
           0x23: 'Safety Discrete I/O Device',
           0x24: 'Fluid Flow Controller',
           0x25: 'CIP Motion Drive',
           0x26: 'CompoNet Repeater',
           0x27: 'Mass Flow Controller, Enhanced',
           0x28: 'CIP Modbus Device',
           0x29: 'CIP Modbus Translator',
           0x2A: 'Safety Analog I/O Device',
           0x2B: 'Generic Device (keyable)',
           0x2C: 'Managed Switch',
           0x32: 'ControlNet Physical Layer Component'}

# List originally came from Wireshark /epan/dissectors/packet-cip.c
vendors = {1: 'Rockwell Automation/Allen-Bradley',
           2: 'Namco Controls Corp.',
           3: 'Honeywell International Inc.',
           4: 'Parker Hannifin Corporation',
           5: 'Rockwell Automation/Reliance Elec.',
           7: 'SMC Corporation',
           8: 'Molex Incorporated',
           9: 'Western Reserve Controls Inc.',
           10: 'Advanced Micro Controls Inc. (AMCI)',
           11: 'ASCO Pneumatic Controls',
           12: 'Banner Engineering Corporation',
           13: 'Belden Wire & Cable Company',
           14: 'Cooper Interconnect',
           16: 'Daniel Woodhead Co.',
           17: 'Dearborn Group Inc.',
           19: 'Helm Instrument Company',
           20: 'Huron Net Works',
           21: 'Belden Deutschland GmbH',
           22: 'Online Development, Inc. (OLDI)',
           23: 'Vorne Industries, Inc.',
           24: 'ODVA',
           26: 'Festo',
           30: 'Unico, Inc.',
           31: 'Ross Controls',
           34: 'Hohner Corp.',
           35: 'Micro Mo Electronics, Inc.',
           36: 'MKS Instruments, Inc.',
           37: 'Yaskawa Electric America formerly Magnetek Drives',
           39: 'AVG Automation (Uticor)',
           40: 'WAGO Corporation',
           41: 'Celerity, Inc.',
           42: 'IMI Norgren Limited',
           43: 'BALLUFF',
           44: 'Yaskawa America, Inc.',
           45: 'Eurotherm by Schneider Electric',
           46: 'ABB, Inc.',
           47: 'Omron Corporation',
           48: 'TURCK',
           49: 'Grayhill Inc.',
           50: 'Real Time Automation',
           52: 'ASCO Numatics',
           53: 'Lutze, Inc.',
           56: 'Softing',
           57: 'Pepperl+Fuchs',
           58: 'Spectrum Controls, Inc.',
           59: 'D.I.P. Inc. MKS Inst.',
           60: 'Applied Motion Products, Inc.',
           61: 'Sencon Inc.',
           62: 'High Country Tek',
           63: 'SWAC Automation Consult GmbH',
           64: 'Clippard Instrument Laboratory',
           68: 'Eaton Electrical',
           71: 'Toshiba Corporation',
           72: 'Control Technology Incorporated',
           73: 'TCS (NZ) Ltd.',
           74: 'Hitachi, Ltd.',
           75: 'ABB Robotics',
           76: 'NKE Corporation',
           77: 'Rockwell Software, Inc.',
           78: 'Escort Memory Systems (A Datalogic Group Co.)',
           79: 'Leviton',
           80: 'Industrial Devices Corporation',
           81: 'IXXAT Automation GmbH',
           82: 'Mitsubishi Electric Automation, Inc.',
           83: 'OPTO 22',
           86: 'Horner Electric',
           87: 'Buerkert Fluid Control Systems',
           89: 'Industrial Indexing Systems, Inc.',
           90: 'HMS Networks',
           91: 'Robicon',
           92: 'Helix Technology (Granville-Phillips)',
           93: 'Arlington Laboratory',
           94: 'Advantech Corporation',
           95: 'Square D Company',
           96: 'Digital Electronics Corp.',
           97: 'Danfoss Drives A/S',
           100: 'AVENTICS',
           101: 'Applied Materials, Inc.',
           102: 'SWCC Showa Cable Systems Co., Ltd.',
           103: 'Pacific Scientific (API Controls Inc.)',
           104: 'Sharp Manufacturing Systems Corporation',
           105: 'Olflex Wire & Cable, Inc.',
           107: 'Unitrode',
           108: 'Beckhoff Automation',
           109: 'National Instruments',
           110: 'Mykrolis Corporation (Millipore)',
           111: 'International Motion Controls Corp.',
           113: 'SEG Kempen GmbH',
           116: 'Temposonics, LLC',
           117: 'Krones, Inc',
           119: 'EXOR Electronic R & D',
           120: 'SIEI S.p.A.',
           121: 'KUKA Deutschland GmbH',
           123: 'SEC (Samsung Electronics Co., Ltd)',
           124: 'Binary Electronics Ltd',
           125: 'Flexible Machine Controls',
           127: 'ABB Inc. (Entrelec)',
           128: 'MAC Valves, Inc.',
           129: 'AUMA Riester GmbH & Co. KG',
           130: 'JTEKT Corporation',
           133: 'Balogh T.A.G., Corporation',
           134: 'TR Electronic',
           135: 'UNIPULSE Corporation',
           138: 'Conxall Corporation Inc.',
           141: 'Kuramo Electric Co., Ltd.',
           142: 'Creative Micro Designs',
           143: 'GE Industrial Systems',
           144: 'Leybold GmbH',
           145: 'Siemens Energy & Automation/Drives',
           146: 'Kodensha Ltd',
           147: 'Motion Engineering, Inc.',
           148: 'Honda Engineering Co., Ltd',
           149: 'EIM Valve Controls',
           150: 'Melec Inc.',
           151: 'Sony Manufacturing Systems Corporation',
           152: 'North American Mfg.',
           153: 'Watlow',
           154: 'Japan Radio Co., Ltd',
           155: 'NADEX Co., Ltd',
           156: 'Ametek Automation & Process Technologies',
           157: 'Facts, Inc.',
           158: 'KVASER AB',
           159: 'IDEC Corporation',
           160: 'Mitsubishi Heavy Industries Ltd',
           161: 'Mitsubishi Electric Corporation',
           162: 'HORIBA STEC, Co., Ltd.',
           163: 'esd electronics gmbh',
           164: 'DAIHEN Corporation',
           165: 'Tyco Valves & Controls/Keystone',
           166: 'EBARA Corporation',
           169: 'Hokuyo Automatic Co., Ltd.',
           170: 'Pyramid Solutions, Inc.',
           171: 'Denso Wave Incorporated',
           172: 'HLS Hard-Line Solutions Inc',
           173: 'Caterpillar, Inc.',
           174: 'PDL Electronics Ltd.',
           176: 'Red Lion',
           177: 'CANON ANELVA Corporation',
           178: 'Toyo Denki Seizo KK',
           179: 'Sanyo Denki Co., Ltd.',
           180: 'Hitachi Metals, Ltd. (formerly Advanced Energy Japan K.K.)',
           181: 'Pilz GmbH & Co KG',
           182: 'Marsh Bellofram-Bellofram PCD Division',
           184: 'M-SYSTEM Co., Ltd.',
           185: 'Nissin Electric Co., Ltd.',
           186: 'Hitachi Metals Ltd.',
           187: 'Oriental Motor Co., Ltd',
           188: 'A&D Company Limited',
           189: 'Phasetronics, Inc.',
           190: 'Cummins Engine Company',
           191: 'Deltron Inc.',
           192: 'Geneer Corporation',
           193: 'Anatol Automation, Inc.',
           196: 'Medar, Inc.',
           197: 'XP Power LLC',
           198: 'Advanced Energy Industries, Inc.',
           200: 'DAIDEN Co., Ltd',
           201: 'CKD Corporation',
           202: 'Toyo Electric Corporation',
           204: 'AuCom Electronics Ltd',
           205: 'Sinfonia Technology Co., Ltd.',
           206: 'Vector Informatik GmbH',
           208: 'Moog Inc.',
           209: 'Contemporary Controls',
           210: 'Tokyo Sokki Kenkyujo Co., Ltd',
           211: 'Schenck Process Group',
           212: 'The Oilgear Company',
           214: 'ASM Japan K.K.',
           215: 'HIRATA Corp.',
           216: 'Panasonic Industrial Devices SUNX Co., Ltd.',
           217: 'Meidensha Corporation',
           218: 'NIDEC SANKYO CORPORATION (Sankyo Seiki Mfg. Co., Ltd)',
           219: 'KAMRO Corp.',
           220: 'NSD Co., Ltd.',
           221: 'EBARA Technologies Inc.',
           224: 'SG Co., Ltd',
           225: 'Vaasa Institute of Technology',
           226: 'MKS Instruments (ENI Technology)',
           227: 'Tateyama System Laboratory Co., Ltd.',
           228: 'QLOG Corporation',
           229: 'Matric Limited Inc.',
           230: 'NSD Corporation',
           232: 'Sumitomo Wiring Systems, Ltd',
           233: 'Group 3 Technology Ltd',
           234: 'CTI Cryogenics',
           235: 'POLSYS CORP',
           236: 'Ampere Inc.',
           238: 'Simplatroll Ltd',
           241: 'Leading Edge Design',
           242: 'Humphrey Products',
           243: 'Schneider Electric',
           244: 'Westlock Controls Corp.',
           245: 'Nihon Weidmuller Co., Ltd',
           246: 'Brooks Instrument',
           248: 'Eaton Industries GmbH (formerly Moeller GmbH)',
           249: 'Varian Vacuum Products',
           250: 'Yokogawa Electric Corporation',
           251: 'Electrical Design Daiyu Co., Ltd',
           252: 'Omron Software Co., Ltd.',
           253: 'EDWARDS',
           254: 'Control Technology Corporation',
           255: 'Bosch Rexroth',
           256: 'TURCK (formerly InterlinkBT)',
           257: 'Control Techniques PLC',
           258: 'Hardy Process Solutions',
           259: 'LS ELECTRIC',
           260: 'E.O.A. Systems Inc.',
           262: 'New Cosmos Electric Co., Ltd.',
           263: 'Sense Sense Eletronica LTDA',
           264: 'Xycom, Inc.',
           265: 'Baldor Electric',
           267: 'Patlite Corporation',
           269: 'Mogami Wire & Cable Corporation',
           270: 'Welding Technology Corporation',
           272: 'Deutschmann Automation GmbH',
           273: 'ICP Panel-Tec Inc.',
           274: 'Bray International, Inc',
           276: 'Status Technologies',
           277: 'Trio Motion Technology ltd',
           278: 'Sherrex Systems Ltd',
           279: 'Adept Technology, Inc.',
           280: 'Spang Power Electronics',
           282: 'Acrosser Technology Co., Ltd',
           283: 'Hilscher GmbH',
           284: 'IMAX Corporation',
           285: 'Electronic Innovation, Inc. (Falter Engineering)',
           286: 'Netlogic Inc.',
           287: 'Bosch Rexroth AG',
           290: 'Murata  Machinery Ltd.',
           291: 'MTT Company Ltd.',
           292: 'Kanematsu Semiconductor Corp.',
           293: 'Takebishi Electric Sales Co.',
           294: 'Tokyo Electron Device Ltd',
           295: 'PFU Limited',
           296: 'Hakko Automation Co., Ltd.',
           297: 'Advanet Inc.',
           298: 'Tokyo Electron Software Technologies Ltd.',
           300: 'Shinagawa Electric Wire Co., Ltd.',
           301: 'Yokogawa M&C Corporation',
           302: 'KONAN Electric Co., Ltd.',
           303: 'Binar Elektronik AB',
           304: 'Furukawa Electric Co.',
           305: 'Cooper Energy Services',
           306: 'Schleicher GmbH & Co.',
           307: 'Hirose Electric Co., Ltd',
           308: 'Western Servo Design Inc.',
           309: 'ProSoft Technology',
           311: 'Towa Shoko Co., Ltd',
           312: 'Kyopal Co., Ltd',
           313: 'Extron Co.',
           314: 'Wieland Electric GmbH',
           315: 'SEW Eurodrive GmbH',
           316: 'Aera Corporation',
           317: 'STA Reutlingen',
           319: 'Fuji Electric Group',
           322: 'ifm electronic gmbh',
           324: 'IDEACOD-Hohner Automation S.A.',
           325: 'CommScope Inc.',
           326: 'Intelligent Platforms, LLC.',
           327: 'Matsushita Electric Industrial Co., Ltd',
           328: 'Okaya Electronics Corporation',
           329: 'KASHIYAMA Industries, Ltd',
           330: 'JVC',
           331: 'Interface Corporation',
           332: 'Grape Systems Inc.',
           335: 'Toshiba IT & Control Systems Corporation',
           336: 'Sanyo Machine Works, Ltd.',
           337: 'Vansco Electronics Ltd.',
           338: 'Dart Container Corp.',
           339: 'Livingston & Co., Inc.',
           340: 'Alfa Laval LKM as',
           341: 'BF ENTRON Ltd. (British Federal)',
           342: 'Bekaert Engineering NV',
           343: 'Ferran  Scientific Inc.',
           344: 'KEBA Industrial Automation GmbH',
           345: 'Endress + Hauser',
           346: 'Lincoln Electric Company',
           347: 'ABB ALSTOM Power UK Ltd. (EGT)',
           348: 'Berger Lahr GmbH',
           350: 'Federal Signal Corp.',
           351: 'Kawasaki Robotics (USA), Inc.',
           352: 'Bently Nevada Corporation',
           354: 'FRABA Posital',
           355: 'Elsag Bailey, Inc.',
           356: 'FANUC Robotics America',
           358: 'Surface Combustion, Inc.',
           360: 'AILES Electronics Ind. Co., Ltd.',
           361: 'Wonderware Corporation',
           362: 'Particle Measuring Systems, Inc.',
           365: 'BITS Co., Ltd',
           366: 'Japan Aviation Electronics Industry Ltd',
           367: 'Keyence Corporation',
           368: 'Kuroda Precision Industries Ltd.',
           369: 'Mitsubishi Electric Semiconductor Application',
           370: 'Nippon Seisen Cable, Ltd.',
           371: 'Omron ASO Co., Ltd',
           372: 'Seiko Seiki Co., Ltd.',
           373: 'Sumitomo Heavy Industries, Ltd.',
           374: 'Tango Computer Service Corporation',
           375: 'Technology Service, Inc.',
           376: 'Toshiba Information Systems (Japan) Corporation',
           377: 'TOSHIBA Schneider Inverter Corporation',
           378: 'Toyooki Kogyo Co., Ltd.',
           379: 'XEBEC',
           380: 'Madison Cable Corporation',
           381: 'Hitati Engineering & Services Co., Ltd',
           382: 'TEM-TECH Lab Co., Ltd',
           383: 'International Laboratory Corporation',
           384: 'Dyadic Systems Co., Ltd.',
           385: 'SETO Electronics Industry Co., Ltd',
           386: 'Tokyo Electron Kyushu Limited',
           387: 'KEI System Co., Ltd',
           389: 'Asahi Engineering Co., Ltd',
           390: 'Contrex Inc.',
           391: 'Paradigm Controls Ltd.',
           393: 'Ohm Electric Co., Ltd.',
           394: 'RKC Instrument Inc.',
           395: 'Suzuki Motor Corporation',
           396: 'Custom Servo Motors Inc.',
           397: 'PACE Control Systems',
           398: 'Selectron Systems AG',
           400: 'LINTEC Co., Ltd.',
           401: 'Hitachi Cable Ltd.',
           402: 'BUSWARE Direct',
           403: 'Eaton Electric B.V. (former Holec Holland N.V.)',
           404: 'VAT Vacuum Valves AG',
           405: 'Omron Robotics and Safety Technologies, Inc.',
           406: 'Alfa Instrumentos Eletronicos Ltda',
           407: 'TWK Elektronik GmbH',
           408: 'ABB Welding Systems AB',
           409: 'BYSTRONIC Maschinen AG',
           410: 'Kimura Electric Co., Ltd',
           411: 'Nissei Plastic Industrial Co., Ltd',
           413: 'Kistler-Morse',
           414: 'Proteus Industries Inc.',
           415: 'IDC Corporation',
           416: 'Nordson Corporation',
           417: 'Rapistan Systems',
           418: 'LP-Elektronik GmbH',
           419: 'GERBI & FASE S.p.A.(Fase Saldatura)',
           420: 'Phoenix Digital Corporation',
           421: 'Z-World Engineering',
           422: 'Honda R&D Co., Ltd.',
           423: 'Bionics Instrument Co., Ltd.',
           424: 'Teknic, Incorporated',
           425: 'R.Stahl, Inc.',
           427: 'Ryco Graphic Manufacturing Inc.',
           428: 'Giddings & Lewis, Inc.',
           429: 'Koganei Corporation',
           431: 'Nichigoh Communication Electric Wire Co., Ltd.',
           433: 'Fujikura Ltd.',
           434: 'AD Link Technology Inc.',
           435: 'Valmet Flow Control Inc (formerly StoneL)',
           436: 'Computer Optical Products, Inc.',
           437: 'CONOS Inc.',
           438: 'Erhardt+Leimer GmbH',
           439: 'UNIQUE Co. Ltd',
           440: 'Roboticsware, Inc.',
           441: 'Nachi Fujikoshi Corporation',
           442: 'Hengstler GmbH',
           443: 'Vacon Plc',
           444: 'SUNNY GIKEN Inc.',
           445: 'Lenze Drive Systems GmbH',
           446: 'CD Systems B.V.',
           447: 'FMT/Aircraft Gate Support Systems AB',
           448: 'Axiomatic Technologies Corp',
           449: 'Embedded System Products, Inc.',
           451: 'Mencom Corporation',
           452: 'Kollmorgen',
           453: 'Panasonic Connect Co., Ltd.',
           454: 'Dengensha Mfg. Co. Ltd.',
           455: 'Quinn Systems Ltd.',
           456: 'Tellima Technology Ltd',
           457: 'MDT, Software',
           458: 'Taiwan Keiso Co., Ltd',
           459: 'Pinnacle Systems',
           460: 'Ascom Hasler Mailing Sys',
           461: 'INSTRUMAR Limited',
           463: 'Navistar International Transportation Corp',
           464: 'TRUMPF Huettinger',
           465: 'OCM Technology Inc.',
           466: 'Professional Supply Inc.',
           467: 'Control Solutions',
           468: 'Baumer IVO GmbH & Co. KG',
           469: 'Worcester Controls Corporation',
           470: 'Pyramid Technical Consultants, Inc.',
           471: 'Eilersen Electric A/S',
           472: 'Apollo Fire Detectors Limited',
           473: 'Avtron Manufacturing, Inc.',
           475: 'Tokyo Keiso Co., Ltd.',
           476: 'Daishowa Swiki Co., Ltd.',
           477: 'Kojima Instruments Inc.',
           478: 'Shimadzu Corporation',
           479: 'Tatsuta Electric Wire & Cable Co., Ltd.',
           480: 'MECS Corporation',
           481: 'Tahara Electric',
           482: 'JTEKT ELECTRONICS CORPORATION',
           483: 'Clever Devices',
           484: 'GCD Hardware & Software GmbH',
           486: 'Miller Electric Mfg Co.',
           487: 'GEA Tuchenhagen GmbH',
           488: 'Riken Keiki Co., Ltd.',
           489: 'Keisokugiken Corporation',
           490: 'Fuji Machine Mfg. Co., Ltd',
           492: 'Nidec-Shimpo Corp.',
           493: 'UTEC Corporation',
           494: 'Sanyo Electric Co. Ltd.',
           497: 'Okano Electric Wire Co. Ltd',
           498: 'Shimaden Co. Ltd.',
           499: 'Teddington Controls Ltd',
           501: 'YASKAWA Europe (formerly VIPA GmbH)',
           502: 'Warwick Manufacturing Group',
           503: 'Danaher Controls',
           506: 'American Science & Engineering',
           507: 'Accutron Controls International Inc.',
           508: 'Norcott Technologies Ltd',
           509: 'TB Woods, Inc',
           510: 'Proportion-Air, Inc.',
           511: 'SICK Stegmann GmbH',
           513: 'Edwards Signaling',
           514: 'Sumitomo Metal Industries, Ltd',
           515: 'Cosmo Instruments Co., Ltd.',
           516: 'Denshosha Co., Ltd.',
           517: 'Kaijo Corp.',
           518: 'Michiproducts Co., Ltd.',
           519: 'Miura Corporation',
           520: 'TG Information Network Co., Ltd.',
           521: 'Fujikin, Inc.',
           522: 'Estic Corp.',
           523: 'GS Hydraulic Sales',
           524: 'Leuze Electronic GmbH & Co. KG',
           525: 'MTE Limited',
           526: 'Hyde Park Electronics, Inc.',
           527: 'Pfeiffer Vacuum GmbH',
           528: 'Cyberlogic Technologies',
           529: 'OKUMA Corporation FA Systems Division',
           531: 'Kokusai Electric Corporation',
           532: 'SHINKO TECHNOS',
           533: 'Itoh Denki Co., Ltd.',
           534: 'Colorado Flow Tech Inc.',
           535: 'Love Controls Division/Dwyer Inst.',
           536: 'Alstom Drives and Controls',
           537: 'The Foxboro Company',
           538: 'Tescom Corporation',
           540: 'Atlas Copco Airpower NV',
           542: 'Autojet Technologies',
           543: 'Prima Electronics S.p.A.',
           544: 'PMA GmbH',
           545: 'Shimafuji Electric Co., Ltd',
           546: 'Oki Electric Industry Co., Ltd',
           547: 'Kyushu Matsushita Electric Co., Ltd',
           548: 'JMACS',
           549: 'Tsuken Electric Ind Co., Ltd',
           550: 'Tamadic Co.',
           551: 'MAATEL SA',
           552: 'OKUMA America',
           553: 'Control Techniques PLC-NA',
           554: 'TPC Wire & Cable',
           555: 'ATI Industrial Automation',
           556: 'Microcontrol (Australia) Pty Ltd',
           557: 'Serra Soldadura, S.A.',
           558: 'Southwest Research Institute',
           559: 'Cabinplant International',
           560: 'Minebea Intec',
           561: 'Comau S.p.A.',
           562: 'Phoenix Contact',
           563: 'Yokogawa MAT Corporation',
           564: 'asahi sangyo co., ltd.',
           565: 'Valcom',
           566: 'Akita Myotoku Ltd.',
           567: 'OBARA Corp.',
           568: 'Suetron Electronic GmbH',
           570: 'Serck Controls Limited',
           571: 'Fairchild Industrial Products Company',
           572: 'ARO Welding Technologies S.A.S.',
           573: 'M2C GmbH',
           574: 'Shin Caterpillar Mitsubishi Ltd.',
           575: 'Santest Co., Ltd.',
           576: 'Cosmotechs Co., Ltd.',
           577: 'Hitachi Electric Systems',
           578: 'Smartscan Ltd',
           579: 'Woodhead Software & Electronics France',
           580: 'Athena Controls, Inc.',
           581: 'Norgren Automation Solutions, LLC (previously Syron Engineering & Manufacturing, Inc.)',
           582: 'Asahi Optical Co., Ltd.',
           583: 'Sansha Electric Mfg. Co.,Ltd.',
           584: 'CKD Nikki Denso Co,. Ltd.',
           585: 'Star Micronics, Co., Ltd.',
           586: 'Ecotecnia Socirtat Corp.',
           587: 'Lenze',
           588: 'West Instruments Limited',
           589: 'LinMot',
           590: 'Delta Computer Systems, Inc.',
           591: 'FANUC CORPORATION',
           592: 'Hearn-Gu Lee',
           593: 'ABB AG',
           594: 'Orion Machinery Co., Ltd.',
           596: 'Wire-Pro, Inc.',
           597: 'Beijing Huakong Technology Co. Ltd.',
           598: 'Yokoyama Shokai Co., Ltd.',
           599: 'Toyogiken Co., Ltd.',
           600: 'Coester Equipamentos Eletronicos Ltda.',
           601: 'Kawasaki Robot',
           602: 'Electroplating Engineers of Japan Ltd.',
           603: 'Robox S.p.a.',
           604: 'Spraying Systems Company',
           605: 'Benshaw Inc.',
           606: 'ZPA-DP A.S.',
           607: 'Wired Rite Systems',
           608: 'Tandis Research, Inc.',
           609: 'SSD Drives GmbH',
           610: 'ULVAC Japan Ltd.',
           611: 'DYNAX Corporation',
           612: 'Nor-Cal Products, Inc.',
           613: 'Aros Electronics AB',
           614: 'Jun-Tech Co., Ltd.',
           615: 'HAN-MI Co. Ltd.',
           616: 'uniNtech (formerly SungGi Internet)',
           617: 'Hae Pyung Electronics Reserch Institute',
           618: 'Milwaukee Electronics',
           619: 'OBERG Industries',
           620: 'Parker Hannifin/Compumotor Division',
           621: 'TECHNO DIGITAL CORPORATION',
           622: 'Network Supply Co., Ltd.',
           623: 'Union Electronics Co., Ltd.',
           624: 'Tritronics Services PM Ltd.',
           625: 'Rockwell Automation/Sprecher+Schuh',
           626: 'Panasonic Corporation/Motor Company',
           627: 'Rolls-Royce Energy Systems, Inc.',
           628: 'JEONGIL INTERCOM CO., LTD',
           629: 'Interroll Engineering GmbH',
           630: 'Hubbell Wiring Device-Kellems (Delaware)',
           631: 'Intelligent Motion Systems',
           633: 'INFICON AG',
           634: 'Hirschmann, a Belden brand',
           635: 'The Siemon Company',
           636: 'YAMAHA Motor Co. Ltd.',
           637: 'aska corporation',
           638: 'Woodhead Connectivity',
           639: 'Trimble AB',
           640: 'Murrelektronik GmbH',
           641: 'Creatrix Labs, Inc.',
           642: 'TopWorx',
           643: 'Kumho Industrial Co., Ltd.',
           644: 'Wind River Systems, Inc.',
           645: 'Bihl + Wiedemann GmbH',
           646: 'Harmonic Drive Systems Inc.',
           647: 'Rikei Corporation',
           648: 'BL Autotec, Ltd.',
           649: 'Hana Information & Technology Co., Ltd.',
           650: 'Seoil Electric Co., Ltd.',
           651: 'Fife Corporation',
           652: 'Shanghai Electrical Apparatus Research Institute',
           653: 'Detector Electronics',
           654: 'Parasense Development Centre',
           657: 'Six Tau S.p.A.',
           658: 'Aucos GmbH',
           659: 'Rotork Controls Ltd.',
           660: 'Automationdirect.com',
           661: 'Thermo BLH',
           662: 'System Controls, Ltd.',
           663: 'Univer S.p.A.',
           664: 'MKS-Tenta Technology',
           665: 'Lika Electronic',
           666: 'Mettler-Toledo',
           667: 'DXL USA Inc.',
           668: 'Rockwell Automation/Entek IRD Intl.',
           669: 'Nippon Otis Elevator Company',
           670: 'Sinano Electric, Co., Ltd.',
           671: 'Sony Manufacturing Systems',
           673: 'CONTEC CO., LTD.',
           674: 'Automated Solutions',
           675: 'Controlweigh',
           677: 'Fincor Electronics',
           678: 'Cognex Corporation',
           679: 'Qualiflow',
           680: 'Weidmuller, Inc.',
           681: 'Morinaga Milk Industry Co., Ltd.',
           682: 'Takagi Industrial Co., Ltd.',
           683: 'Wittenstein SE',
           684: 'Sena Technologies, Inc.',
           686: 'SPX Flow Technology Germany GmbH',
           687: 'Creator Teknisk Utvedkling AB',
           689: 'Mibu Denki Industrial Co., Ltd.',
           690: 'Takamastsu Machineer Section',
           691: 'Littelfuse',
           693: 'Holjeron',
           694: 'Pfeiffer Vacuum SAS',
           695: 'Taesan LCD Co., Ltd.',
           696: 'POSCON',
           697: 'VMIC',
           698: 'Matsushita Electric Works, Ltd.',
           699: 'IAI Corporation',
           700: 'Horst GmbH',
           701: 'MicroControl GmbH & Co.',
           702: 'Leine & Linde AB',
           704: 'EC Elettronica Srl',
           705: 'VIT Software HB',
           706: 'Bronkhorst High-Tech B.V.',
           707: 'Optex Co., Ltd.',
           708: 'Yosio Electronic Co.',
           709: 'Terasaki Electric Co., Ltd.',
           710: 'Sodick Co., Ltd.',
           711: 'MTS Systems Corporation-Automation Division',
           712: 'Mesa Systemtechnik',
           713: 'SHIN HO SYSTEM Co., Ltd.',
           714: 'Goyo Electronics Co, Ltd.',
           715: 'Loreme',
           716: 'SAB Brockskes GmbH & Co. KG',
           717: 'Trumpf Laser GmbH',
           718: 'Niigata Electronic Instruments Co., Ltd.',
           719: 'Yokogawa Digital Computer Corporation',
           720: 'O.N. Electronic Co., Ltd.',
           721: 'Industrial Control Communication, Inc.',
           722: 'ABB, Inc.',
           723: 'ElectroWave USA, Inc.',
           724: 'Industrial Network Controls, LLC',
           725: 'KDT Systems Co., Ltd.',
           726: 'SEFA Technology Inc.',
           727: 'Nippon POP Rivets and Fasteners Ltd.',
           728: 'Yamato Scale Co., Ltd.',
           729: 'Zener Electric',
           730: 'GSE Scale Systems',
           731: 'ISAS (Integrated Switchgear & Sys. Pty Ltd)',
           732: 'Beta LaserMike Limited',
           733: 'TOEI Electric Co., Ltd.',
           734: 'Hakko Electronics Co., Ltd',
           736: 'RFID, Inc.',
           737: 'Adwin Corporation',
           738: 'Osaka Vacuum, Ltd.',
           739: 'A-Kyung Motion, Inc.',
           740: 'Camozzi Automation spa',
           741: 'Crevis Co., LTD',
           742: 'Rice Lake Weighing Systems',
           743: 'Linux Network Services',
           744: 'KEB Automation KG',
           745: 'Hagiwara Electric Co., Ltd.',
           746: 'Glass Inc. International',
           748: 'DVT Corporation',
           749: 'Woodward Governor',
           750: 'Mosaic Systems, Inc.',
           751: 'Laserline GmbH',
           752: 'COM-TEC, Inc.',
           753: 'Weed Instrument',
           754: 'Prof-face European Technology Center',
           755: 'Fuji Automation Co., Ltd.',
           756: 'Matsutame Co., Ltd.',
           757: 'Hitachi Via Mechanics, Ltd.',
           758: 'Dainippon Screen Mfg. Co. Ltd.',
           759: 'FLS Automation A/S',
           760: 'ABB Stotz Kontakt GmbH',
           761: 'Technical Marine Service',
           762: 'Advanced Automation Associates, Inc.',
           763: 'Baumer Ident GmbH',
           764: 'Tsubakimoto Chain Co.',
           766: 'Furukawa Co., Ltd.',
           767: 'Active Power',
           768: 'CSIRO Mining Automation',
           769: 'Matrix Integrated Systems',
           770: 'Digitronic Automationsanlagen GmbH',
           771: 'SICK STEGMANN Inc.',
           772: 'TAE-Antriebstechnik GmbH',
           773: 'Electronic Solutions',
           774: 'Rocon L.L.C.',
           775: 'Dijitized Communications Inc.',
           776: 'Asahi Organic Chemicals Industry Co., Ltd.',
           777: 'Hodensha',
           778: 'HARTING, Inc. of North America',
           779: 'Kuebler GmbH',
           780: 'Yamatake Corporation',
           781: 'JEOL',
           782: 'Yamatake Industrial Systems Co., Ltd.',
           783: 'HAEHNE Elektronische Messgerate GmbH',
           784: 'Ci Technologies Pty Ltd (for Pelamos Industries)',
           785: 'N. SCHLUMBERGER & CIE',
           786: 'Teijin Seiki Co., Ltd.',
           787: 'DAIKIN Industries, Ltd',
           788: 'RyuSyo Industrial Co., Ltd.',
           789: 'SAGINOMIYA SEISAKUSHO, INC.',
           790: 'Seishin Engineering Co., Ltd.',
           791: 'Japan Support System Ltd.',
           792: 'Decsys',
           793: 'Metronix Messgerate u. Elektronik GmbH',
           794: 'ROPEX Industrie - Elektronik GmbH',
           795: 'Vaccon Company, Inc.',
           796: 'Siemens Industry, Inc.',
           797: 'Ten X Technology, Inc.',
           798: 'TE Connectivity',
           799: 'Delta Electronics, Inc.',
           800: 'Denker',
           801: 'Autonics Corporation',
           802: 'JFE Electronic Engineering Pty. Ltd.',
           804: 'Electro-Sensors, Inc.',
           805: 'Digi International, Inc.',
           806: 'Texas Instruments',
           807: 'ADTEC Plasma Technology Co., Ltd',
           808: 'SICK AG',
           809: 'Ethernet Peripherals, Inc.',
           810: 'Animatics Corporation',
           812: 'Process Control Corporation',
           813: 'SystemV. Inc.',
           814: 'Danaher Motion SRL',
           815: 'SHINKAWA Sensor Technology, Inc.',
           816: 'Tesch GmbH & Co. KG',
           818: 'Trend Controls Systems Ltd.',
           819: 'Guangzhou ZHIYUAN Electronic Co., Ltd.',
           820: 'Mykrolis Corporation',
           821: 'Bethlehem Steel Corporation',
           822: 'KK ICP',
           823: 'Takemoto Denki Corporation',
           824: 'Montalvo Corporation',
           826: 'LEONI Special Cables GmbH',
           828: 'ONO SOKKI CO.,LTD.',
           829: 'RS Automation Co., Ltd.',
           830: 'SHINDENGEN ELECTRIC MFG. CO. LTD',
           831: 'Origin Electric Co. Ltd.',
           832: 'Quest Technical Solutions, Inc.',
           833: 'LS Cable',
           834: 'NORD Electronic DRIVESYSTEMS GmbH',
           835: 'Northwire Inc.',
           836: 'Engel Elektroantriebe GmbH',
           837: 'The Stanley Works',
           838: 'Celesco Transducer Products, Inc.',
           839: 'Chugoku Electric Wire and Cable Co.',
           840: 'Kongsberg Simrad AS',
           841: 'Panduit Corporation',
           842: 'Spellman High Voltage Electronics Corp.',
           843: 'Kokusai Electric Alpha Co., Ltd.',
           844: 'Brooks Automation, Inc.',
           845: 'ANYWIRE CORPORATION',
           846: 'Honda Electronics Co. Ltd',
           847: 'REO AG',
           848: 'Heraeus Noblelight Fusion UV Inc.',
           849: 'ASI Advanced Semiconductor Instruments GmbH',
           850: 'Datalogic, Inc.',
           851: 'SoftPLC Corporation',
           852: 'Dynisco Instruments LLC',
           853: 'WEG',
           854: 'Teledyne LeCroy (formerly Frontline Test Equipment)',
           855: 'Tamagawa Seiki Co., Ltd.',
           856: 'Multi Computing Co., Ltd.',
           857: 'RVSI',
           858: 'Commercial Timesharing Inc.',
           859: 'Tennessee Rand Automation LLC',
           860: 'Wacogiken Co., Ltd',
           861: 'Reflex Integration Inc.',
           862: 'Siemens AG, A&D PI Flow Instruments',
           863: 'G. Bachmann Electronic GmbH',
           864: 'Entegris, Inc.',
           865: 'Schweitzer Engineering Laboratories',
           866: 'ATR Industrie-Elektronik GmbH Co.',
           867: 'PLASMATECH Co., Ltd',
           869: 'GEMUE GmbH & Co. KG',
           870: 'Alcorn McBride Inc.',
           871: 'MORI SEIKI CO., LTD',
           872: 'NodeTech Systems Ltd',
           873: 'Emhart Teknologies',
           874: 'Cervis, Inc.',
           875: 'MSA Safety',
           876: 'NEDAP Power Supplies',
           877: 'Nippon Sanso Corporation',
           878: 'Mitomi Giken Co., Ltd.',
           879: 'PULS GmbH',
           881: 'Japan Control Engineering Ltd',
           882: 'Embedded Systems Korea (Former Zues Emtek Co Ltd.)',
           883: 'Automa SRL',
           884: 'Harms+Wende GmbH & Co KG',
           885: 'R. STAHL',
           886: 'Microwave Data Systems',
           887: 'Bernecker + Rainer Industrie-Elektronik GmbH',
           888: 'Hiprom Technologies',
           890: 'Nitta Corporation',
           891: 'Kontron Modular Computers GmbH',
           892: 'Marlin Controls',
           893: 'Elcis Encoder s.r.l.',
           894: 'Acromag, Inc.',
           895: 'Avery Weigh-Tronix',
           899: 'Practicon Ltd.',
           900: 'SCHUNK GmbH & Co. KG',
           901: 'MYNAH Technologies',
           902: 'Defontaine Groupe',
           903: 'Emerson Process Management Power & Water Solutions',
           904: 'F.A. Elec',
           905: 'Hottinger Baldwin Messtechnik GmbH',
           906: 'Teledyne DALSA',
           907: 'London Electronics Ltd.',
           908: 'HSD SpA',
           909: 'Pepperl+Fuchs Comtrol',
           910: 'TEAM, S.A. (Tecnica Electronica de Automatismo Y Medida)',
           911: 'Regulateurs Europa Ltd',
           914: 'Micro Motion',
           915: 'Eckelmann AG',
           916: 'Hanyoung Nux',
           917: 'CFT Ransburg Japan KK',
           918: 'Kun Hung Electric Co. Ltd.',
           919: 'Brimos wegbebakening b.v.',
           920: 'NITTO SEIKO CO., LTD.',
           921: 'Datasensing S.r.l.',
           922: 'Yamazaki Machinery Works',
           923: 'Schmidt Technology',
           924: 'Parker Hannifin SpA (SBC Division)',
           925: 'HIMA Paul Hildebrandt GmbH',
           926: 'RivaTek, Inc.',
           927: 'Misumi Corporation',
           928: 'GE Multilin',
           929: 'Measurement Computing Corporation',
           930: 'Jetter AG',
           931: 'Tokyo Electronics Systems Corporation',
           932: 'Togami Electric Mfg. Co., Ltd.',
           933: 'HK Systems',
           934: 'CDA Systems Ltd.',
           935: 'Aerotech Inc.',
           936: 'JVL A/S',
           937: 'NovaTech Process Solutions LLC',
           939: 'Cisco Systems, Inc.',
           940: 'Grid Connect',
           941: 'ITW Automotive Finishing',
           942: 'HanYang System',
           943: 'ABB K.K. Technical Center',
           944: 'Taiyo Cable (Dongguan) Co., Ltd.',
           946: 'SEREN IPS INC',
           947: 'Belden',
           948: 'ControlNet International',
           949: 'Gefran S.P.A.',
           950: 'ABB (Jokab Safety)',
           951: 'SUMITA OPTICAL GLASS, INC.',
           952: 'Biffi Italia srl',
           953: 'Beck IPC GmbH',
           954: 'Copley Controls',
           955: 'Fagor Automation S. Coop.',
           956: 'DARCOM',
           957: 'Frick Controls (div. of York International)',
           958: 'SymCom, Inc.',
           959: 'Infranor',
           960: 'Kyosan Electric Mfg',
           961: 'Varian Vacuum Technologies',
           962: 'Messung Systems',
           963: 'Xantrex Technology, Inc.',
           964: 'StarThis Inc.',
           965: 'NF Chiyoda Electronics Co., Ltd.',
           966: 'Flowserve Corporation',
           967: 'Spyder Controls Corp.',
           968: 'IBA AG',
           969: 'SHIMOHIRA ELECTRIC MFG.CO.,LTD',
           971: 'Siemens L&A',
           972: 'Eaton Automation GmbH (formerly Micro Innovation)',
           973: 'Switchgear & Instrumentation',
           974: 'Pre-Tech Co., Ltd.',
           975: 'National Semiconductor',
           976: 'Invensys Operations Management',
           977: 'Ametek HDR Power Systems',
           979: 'TETRA-K Corporation',
           980: 'C&M Corporation',
           981: 'Siempelkamp Maschinen',
           983: 'Daifuku Co., Ltd',
           984: 'Electro-Matic Products Inc.',
           985: 'BUSSAN MICROELECTRONICS CORP.',
           986: 'ELAU AG',
           987: 'Hetronic USA',
           988: 'NIIGATA POWER SYSTEMS Co., Ltd.',
           989: 'Software Horizons Inc.',
           990: 'B3 Systems, Inc.',
           991: 'Moxa, Inc.',
           993: 'S4 Integration',
           994: 'Elettro Stemi S.R.L.',
           995: 'AquaSensors',
           996: 'Ifak System GmbH',
           997: 'SANKEI MANUFACTURING Co.,LTD.',
           998: 'Emerson Network Power Co., Ltd.',
           999: 'Fairmount Automation, Inc.',
           1000: 'Bird Electronic Corporation',
           1001: 'Nabtesco Corporation',
           1002: 'AGM Electronics, Inc.',
           1003: 'ARCX Inc.',
           1004: 'DELTA I/O Co.',
           1005: 'Chun IL Electric Ind. Co.',
           1006: 'N-Tron Corporation, a Red Lion Company',
           1007: 'Nippon Pneumatics/Fludics System CO.,LTD.',
           1008: 'DDK Ltd.',
           1009: 'Seiko Epson Corporation',
           1010: 'halstrup-walcher GmbH',
           1011: 'ITT',
           1012: 'Ground Fault Systems bv',
           1013: 'Scolari Engineering S.p.A.',
           1014: 'Vialis Traffic bv',
           1015: 'Weidmueller Group',
           1016: 'Shanghai Sibotech Automation Co. Ltd',
           1017: 'AEG Power Solutions GmbH',
           1018: 'Komatsu Electronics Inc.',
           1019: 'Souriau',
           1020: 'Baumuller Chicago Corp.',
           1021: 'J. Schmalz GmbH',
           1022: 'SEN Corporation',
           1023: 'Korenix Technology Co. Ltd',
           1024: 'Cooper Power Tools',
           1025: 'INNOBIS',
           1026: 'Shinho System',
           1027: 'Xm Services Ltd.',
           1028: 'KVC Co., Ltd.',
           1029: 'Sanyu Seiki Co., Ltd.',
           1030: 'TuxPLC',
           1031: 'Northern Network Solutions',
           1032: 'Converteam GmbH',
           1033: 'Symbol Technologies',
           1034: 'S-TEAM Lab',
           1035: 'Maguire Products, Inc.',
           1036: 'AC&T',
           1037: 'MITSUBISHI HEAVY INDUSTRIES, LTD. KOBE SHIPYARD & MACHINERY WORKS',
           1038: 'Hurletron Inc.',
           1039: 'Chunichi Denshi Co., Ltd',
           1040: 'Cardinal Scale Mfg. Co.',
           1041: 'METZ CONNECT USA Inc.',
           1042: 'Base2',
           1043: 'ASRC Aerospace',
           1044: 'Beijing Stone Automation',
           1045: 'Changshu Switchgear Manufacture Ltd.',
           1046: 'METRONIX Corp.',
           1047: 'WIT',
           1048: 'ORMEC Systems Corp.',
           1049: 'ASATech (China) Inc.',
           1050: 'Controlled Systems Limited',
           1051: 'Mitsubishi Heavy Ind. Digital System Co., Ltd. (M.H.I.)',
           1052: 'Electrogrip',
           1053: 'TDS Automation',
           1054: 'T&C Power Conversion, Inc.',
           1055: 'Robostar Co., Ltd',
           1056: 'Scancon A/S',
           1057: 'Haas Automation, Inc.',
           1058: 'Eshed Technology',
           1059: 'Delta Electronic Inc.',
           1060: 'Innovasic',
           1061: 'SoftDEL Systems Limited',
           1062: 'FiberFin, Inc.',
           1063: 'Nicollet Technologies Corp.',
           1064: 'B.F. Systems',
           1065: 'Empire Wire and Supply LLC',
           1066: 'ENDO KOGYO CO., LTD',
           1067: 'Elmo Motion Control LTD',
           1069: 'Asahi Keiki Co., Ltd.',
           1070: 'Joy Mining Machinery',
           1071: 'MPM Engineering Ltd.',
           1072: 'Wolke Inks & Printers GmbH',
           1073: 'Mitsubishi Electric Engineering Co., Ltd.',
           1074: 'COMET AG',
           1075: 'Real Time Objects & Systems, LLC',
           1076: 'MISCO Refractometer',
           1077: 'JT Engineering Inc.',
           1078: 'Automated Packing Systems',
           1079: 'Niobrara R&D Corp.',
           1080: 'Garmin Ltd.',
           1081: 'Japan Mobile Platform Co., Ltd',
           1082: 'Advosol Inc.',
           1083: 'ABB Global Services Limited',
           1084: 'Sciemetric Instruments Inc.',
           1085: 'Tata Elxsi Ltd.',
           1086: 'Mechatronics Co.,Ltd',
           1087: 'Cooper Bussmann',
           1088: 'Trinite Automatisering B.V.',
           1089: 'Peek Traffic B.V.',
           1090: 'Acrison, Inc.',
           1091: 'Applied Robotics, Inc.',
           1092: 'FireBus LLC',
           1093: 'Sevenstar',
           1094: 'Magnetek',
           1095: 'Omron Microscan Systems, Inc.',
           1096: 'Air Water Inc.',
           1097: 'SensoPart Industriesensorik GmbH',
           1098: 'Tiefenbach Control Systems GmbH',
           1099: 'INOXPA S.A',
           1100: 'Zurich University of Applied Sciences',
           1101: 'Ethernet Direct',
           1102: 'GSI-Micro-E Systems',
           1103: 'S-Net Automation Co., Ltd.',
           1104: 'Power Electronics S.L.',
           1105: 'Renesas Electronics',
           1106: 'NSWCCD-SSES',
           1107: 'Porter Engineering Ltd.',
           1108: 'Meggitt Airdynamics, Inc.',
           1109: 'Inductive Automation',
           1110: 'Neural ID',
           1111: 'EEPod LLC',
           1112: 'Hitachi Industrial Equipment Systems Co.,Ltd.',
           1113: 'Salem Automation',
           1114: 'port GmbH',
           1115: 'B & PLUS',
           1116: 'Graco Inc.',
           1117: 'Altera Corporation',
           1118: 'Technology Brewing Corporation',
           1121: 'CSE Servelec',
           1124: 'Fluke Networks',
           1125: 'Tetra Pak Packaging Solutions SpA',
           1126: 'Racine Federated, Inc.',
           1127: 'Pureron Japan Co., Ltd.',
           1130: 'Brother Industries, Ltd.',
           1132: 'Leroy Automation',
           1134: 'THK Co., Ltd.',
           1137: 'TR-Electronic GmbH',
           1138: 'ASCON S.p.A.',
           1139: 'Toledo do Brasil Industria de Balancas Ltda.',
           1140: 'Caterpillar Global Mining Europe GmbH',
           1141: 'Emerson Process Management Valve Automation',
           1142: 'Alstom Transport',
           1144: 'Matrox Electronic Systems',
           1145: 'Littelfuse',
           1146: 'PLASMART, Inc.',
           1147: 'Miyachi Corporation',
           1150: 'Promess Incorporated',
           1151: 'COPA-DATA GmbH',
           1152: 'Precision Engine Controls Corporation',
           1153: 'Alga Automacao e controle LTDA',
           1154: 'Lapp Group',
           1155: 'ICES',
           1156: 'Philips Lighting bv',
           1157: 'Aseptomag AG',
           1158: 'ARC Informatique',
           1159: 'Hesmor GmbH',
           1160: 'Kobe Steel, Ltd.',
           1161: 'FLIR Systems',
           1162: 'Xcelgo A/S',
           1163: 'STRATON AUTOMATION',
           1164: 'Zypcom, Inc.',
           1165: 'Swagelok',
           1166: 'Elspec',
           1167: 'ITT Water & Wastewater AB',
           1168: 'Kunbus GmbH Industrial Communication',
           1170: 'Performance Controls, Inc.',
           1171: 'ACS Motion Control, Ltd.',
           1173: 'IStar Technology Limited',
           1174: 'Alicat Scientific, Inc.',
           1176: 'ADFweb.com SRL',
           1177: 'Tata Consultancy Services Limited',
           1178: 'CXR Ltd.',
           1179: 'Vishay Nobel AB',
           1181: 'Emerson - SolaHD',
           1182: 'Endress+Hauser',
           1183: 'Bartec GmbH',
           1185: 'AccuSentry, Inc.',
           1186: 'Curtiss Wright - Exlar Actuator Solutions',
           1187: 'ILS Technology',
           1188: 'Control Concepts, Inc.',
           1190: 'Procon Engineering A Division of National Oilwell Varco UK Ltd',
           1191: 'Hermary',
           1192: 'Q-Lambda',
           1194: 'VAMP Ltd',
           1195: 'FlexLink',
           1196: 'Office FA.com Co., Ltd.',
           1197: 'SPMC (Changzhou) Co. Ltd.',
           1198: 'Anton Paar GmbH',
           1199: 'Zhuzhou CSR Times Electric Co., Ltd.',
           1200: 'DeStaCo',
           1201: 'Synrad, Inc',
           1202: 'Bonfiglioli Vectron GmbH',
           1203: 'Pivotal Systems',
           1204: 'TKSCT',
           1205: 'Randy Nuernberger',
           1206: 'CENTRALP',
           1207: 'Tengen Group',
           1208: 'OES, Inc.',
           1209: 'Actel Corporation',
           1210: 'Monaghan Engineering, Inc.',
           1211: 'wenglor sensoric gmbh',
           1212: 'HSA Systems',
           1213: 'MKP Co., Ltd.',
           1214: 'Tappan Wire and Cable',
           1215: 'Heinzmann GmbH & Co. KG',
           1216: 'Process Automation International Ltd.',
           1217: 'Secure Crossing',
           1218: 'SMA Railway Technology GmbH',
           1219: 'FMS Force Measuring Systems AG',
           1220: 'ABT Endustri Enerji Sistemleri Sanayi Tic. Ltd. Sti.',
           1221: 'MagneMotion Inc.',
           1222: 'STS Co., Ltd.',
           1223: 'Knorr-Bremse Espana, S.A. - Merak Division',
           1224: 'ABOUNDI, Inc.',
           1225: 'Rosemount Inc.',
           1226: 'GEA FES, Inc.',
           1227: 'TMG Technologie und Engineering GmbH',
           1228: 'embeX GmbH',
           1229: 'GH Electrotermia, S.A.',
           1230: 'Tolomatic',
           1231: 'Dukane',
           1232: 'Elco (Tian Jin) Electronics Co., Ltd.',
           1233: 'Jacobs Automation',
           1234: 'Noda Radio Frequency Technologies Co., Ltd.',
           1235: 'MSC Tuttlingen GmbH',
           1236: 'Hitachi Cable Manchester',
           1237: 'ACOREL SAS',
           1238: 'Global Engineering Solutions Co., Ltd.',
           1239: 'ALTE Transportation, S.L.',
           1240: 'Penko Engineering B.V.',
           1241: 'Z-Tec Automation Systems Inc.',
           1242: 'ENTRON Controls LLC',
           1243: 'Johannes Huebner Fabrik Elektrischer Maschinen GmbH',
           1244: 'RF IDeas, Inc.',
           1245: 'Pentronic AB',
           1246: 'Atlas Copco IAS GmbH',
           1247: 'TDK-Lambda',
           1250: 'Altronic LLC',
           1251: 'Siemens AG',
           1252: 'Liebherr Transportation Systems GmbH & Co KG',
           1256: 'LMI Technologies',
           1261: 'CEPHALOS Automatisierung mbH',
           1265: 'Quabbin Wire & Cable Co., Inc.',
           1268: 'HORIBA Precision Instruments (Beijing) Co.,Ltd.',
           1270: 'Rovema GmbH',
           1272: 'IEP GmbH',
           1277: 'Control Chief Corporation',
           1282: 'PRIMES GmbH',
           1283: 'Branson Ultrasonics',
           1284: 'DEIF A/S',
           1285: 'CODESYS GmbH',
           1287: 'Smarteye Corporation',
           1288: 'Shibaura Machine',
           1289: 'HMS/BU Ewon',
           1290: 'OFS',
           1291: 'KROHNE',
           1295: 'Kistler Instrumente AG',
           1301: 'Xylem Analytics Germany GmbH',
           1302: 'Lenord, Bauer & Co. GmbH',
           1303: 'Carlo Gavazzi Controls',
           1304: 'Faiveley Transport',
           1306: 'Sensia LLC',
           1307: 'Kepware Technologies',
           1308: 'duagon AG',
           1310: 'Xylem Water Solutions',
           1311: 'Automation Professionals, LLC',
           1313: 'CEIA SpA',
           1315: 'Alphagate Automatisierungstechnik GmbH',
           1316: 'Mecco Partners, LLC',
           1317: 'LAP GmbH Laser Applikationen',
           1318: 'ABB S.p.A. - SACE Division',
           1321: 'C.E. Electronics, Inc.',
           1322: 'Thermo Ramsey Inc., a part of Thermo Fisher Scientific',
           1323: 'Helmholz GmbH & Co. KG',
           1324: 'EUCHNER GmbH + Co. KG',
           1325: 'AMKmotion',
           1326: 'Badger Meter',
           1328: 'Fisher-Rosemount Systems, Inc. doing business as Process Systems & Solutions',
           1329: 'Conductix-Wampfler Automation GmbH',
           1330: 'Fairbanks Scales, Inc.',
           1331: 'Imperx, Inc.',
           1332: 'FRONIUS International GmbH',
           1333: 'Hoffman Enclosures',
           1334: 'Elecsys Corporation',
           1336: 'RACO Manufacturing and Engineering',
           1337: 'Hein Lanz Industrial Tech.',
           1338: 'Codenomicon',
           1339: 'SABO Elektronik GmbH',
           1341: 'Sensirion AG',
           1342: 'SIKO GmbH',
           1344: 'GRUNDFOS',
           1346: 'Beijer Electronics Products AB',
           1348: 'AIMCO',
           1350: 'Coval',
           1351: 'Powell Industries',
           1353: 'IPDisplays',
           1354: 'SCAIME SAS',
           1355: 'Metal Work SpA',
           1356: 'Telsonic AG',
           1358: 'Hauch & Bach ApS',
           1359: 'Pago AG',
           1360: 'ULTIMATE Europe Transportation Equipment GmbH',
           1362: 'FW Murphy Production Controls, LLC',
           1363: 'Lake Cable LLC',
           1368: 'Nanotec Electronic GmbH & Co. KG',
           1369: 'SAMWON ACT Co., Ltd.',
           1370: 'Aparian Inc.',
           1371: 'Cosys Inc.',
           1372: 'Insight Automation Inc.',
           1374: 'FASTECH',
           1375: 'K.A. Schmersal GmbH & Co. KG',
           1378: 'SEIDENSHA ELECTRONICS CO., LTD',
           1380: 'Don Electronics Ltd',
           1381: 'burster gmbh & co kg',
           1382: 'Unitronics (1989) (RG) LTD',
           1383: 'OEM Technology Solutions',
           1384: 'Allied Motion',
           1386: 'Dengensha TOA',
           1387: 'Systec Systemtechnik und Industrieautomation GmbH',
           1389: 'Jenny Science AG',
           1390: 'Baumer Optronic GmbH',
           1391: 'Invertek Drives Ltd',
           1392: 'High Grade Controls Corporation',
           1396: 'Actia Systems',
           1398: 'Beijing Tianma Intelligent Control Technology Co., Ltd',
           1399: 'Universal Robots A/S',
           1401: 'Dialight',
           1402: 'E-T-A Elektrotechnische Apparate GmbH',
           1403: 'Kemppi Oy',
           1405: 'ORing Industrial Networking Corp.',
           1408: 'ELAP S.R.L.',
           1409: 'Applied Mining Technologies',
           1410: 'KITZ SCT Corporation',
           1411: 'VTEX Corporation',
           1412: 'ESYSE GmbH Embedded Systems Engineering',
           1413: 'Automation Controls',
           1415: 'Cincinnati Test Systems',
           1417: 'Zumbach Electronics Corp.',
           1418: 'Emerson Automation Solutions',
           1419: 'CCS Inc.',
           1420: 'Videojet, Inc.',
           1421: 'Zebra Technologies',
           1422: 'ANRITSU CORPORATION',
           1423: 'Dimetix AG',
           1424: 'General Measure (China)',
           1425: 'Fortress Interlocks',
           1427: 'Task Force Tips',
           1428: 'SERVO-ROBOT INC.',
           1429: 'Flow Devices and Systems, Inc.',
           1430: 'nLIGHT, Inc.',
           1431: 'Microchip Technology Inc.',
           1434: 'Accutron Instruments Inc.',
           1435: 'Kaeser Kompressoren SE',
           1437: 'Coherix, Inc.',
           1438: 'FLSmidth A/S',
           1440: 'Cole-Parmer Instrument Company',
           1441: 'Wachendorff Automation GmbH & Co., KG',
           1442: 'SMAC Moving Coil Actuators',
           1444: 'PushCorp, Inc.',
           1445: 'Fluke Process Instruments GmbH',
           1446: 'Mini Motor S.p.a',
           1447: 'I-CON Industry Tech.',
           1448: 'Grace Technologies',
           1449: 'Zaxis Inc.',
           1450: 'Lumasense Technologies',
           1451: 'Domino Printing',
           1454: 'Altus Sistemas de Automacao S.A.',
           1456: 'InterTech Development Company',
           1458: 'Perle Systems Limited',
           1459: 'Utthunga Technologies Private Limited',
           1461: 'WIPOTEC GmbH',
           1462: 'Atos spa',
           1463: 'Solartron Metrology LTD',
           1465: 'Analog Devices',
           1466: 'Power Electronics International, Inc.',
           1468: 'Campbell Wrapper Corporation',
           1469: 'Herkules-Resotec Elektronik GmbH',
           1470: 'aignep spa',
           1472: 'PMV Automation AB',
           1474: 'ProTec Dynatronix LLC dba Dynatronix',
           1476: 'Bitronics, LLC.',
           1477: 'Delta Tau Data Systems',
           1479: 'AUTOSOL',
           1480: 'ADB Safegate',
           1483: 'Artis GmbH',
           1484: 'REJ Co., LTD',
           1485: 'Vanderlande',
           1486: 'Packet Power',
           1487: 'ima-tec gmbh',
           1488: 'Vision Automation A/S',
           1489: 'PROCENTEC BV',
           1490: 'HETRONIK GmbH',
           1491: 'Lanmark Controls Inc.',
           1493: 'flexlog GmbH',
           1494: 'YUCHANGTECH',
           1495: 'Dynapower Company',
           1496: 'TAKIKAWA ENGINEERING',
           1497: 'Ingersoll Rand',
           1498: 'ASA-RT s.r.l',
           1499: 'TRUMPF Schweiz AG',
           1501: 'Rinstrum',
           1504: 'BlueBotics SA',
           1505: 'Dynapar Corporation',
           1506: 'Blum-Novotest',
           1507: 'CIMON',
           1509: 'Rethink Robotics GmbH',
           1510: 'Ingeteam Power Technology S. A.',
           1511: 'TOSEI ENGINEERING CORP.',
           1512: 'SAMSON AG',
           1513: 'TGW Mechanics GmbH',
           1514: 'Diatrend Corporation',
           1516: 'VAHLE Automation GmbH',
           1517: 'JSL Technology Co.,Ltd.',
           1518: 'NetTechnix E&P GmbH',
           1520: 'Tecweigh',
           1521: 'IVEK Corporation',
           1523: 'AQ M-TECH AB',
           1524: 'Rexnord Industries LLC',
           1526: 'OPTEX FA Co., Ltd',
           1527: 'Volktek Corporation',
           1528: 'INGENIA',
           1530: 'Analytical Technology, Inc.',
           1531: 'Columbus McKinnon Corporation',
           1532: 'HBC-radiomatic GmbH',
           1533: 'Leonton Technologies',
           1534: 'Mitsubishi Electric India Pvt. Ltd.',
           1535: 'FOBA Laser, ALLTEC GmbH',
           1536: 'Leakmaster Inc',
           1537: 'Buhler AG',
           1538: 'LINAK Denmark A/S',
           1540: 'SIEB & MEYER AG',
           1542: 'Watson-Marlow Ltd',
           1543: 'ABB Switzerland Ltd - Low Voltage Products',
           1545: 'adphos',
           1546: 'Hangzhou Hikrobot Technology Co., Ltd.',
           1547: 'TOSS GmbH & Co. KG',
           1548: 'Solar Turbines Incorporated',
           1551: 'Carlo Gavazzi Industri',
           1552: 'Nippon Gear',
           1553: 'OSIsoft, LLC',
           1554: 'Rinco Ultrasonics AG',
           1557: 'Mitutoyo',
           1558: 'swisca',
           1559: 'Micro-Epsilon Messtechnik GmbH & Co. KG',
           1560: 'AMADA MIYACHI AMERICA',
           1561: 'Taihan Electric Wire Co., Ltd.',
           1562: 'JANOME Corporation',
           1563: 'ISHIDA',
           1564: 'NAKANISHI INC.',
           1565: 'Mecademic Inc.',
           1567: 'Sigma (NSW) PTY LTD',
           1568: 'Hammond Power Solutions Inc.-Mesta',
           1571: 'TRIDIMEO',
           1572: 'Motortronics UK Ltd.',
           1573: 'Doosan Robotics',
           1574: 'ADVANCED Motion Controls',
           1575: 'OnRobot A/S',
           1577: 'Oetiker',
           1578: 'SICK OPTEX',
           1581: 'Kahler Automation Corporation',
           1582: 'Accuenergy (Canada) Inc.',
           1583: 'TCI, LLC - An Allied Motion Company',
           1584: 'Sun Automation',
           1585: 'READY Robotics Corporation',
           1586: 'PEM, Power Eng & Mfg., Inc',
           1587: 'Duerr Somac GmbH',
           1591: 'RICOH Industrial Solutions Inc.',
           1592: 'Shanghai Junqian Sensing Technology Co. Ltd.',
           1593: 'Knick Elektronische Messgeraete GmbH & Co. KG',
           1594: 'Magnescale. Co., Ltd.',
           1596: 'Weintek Labs., Inc',
           1597: 'Sherpa, Inc.',
           1598: 'Inspekto A.M.V LTD',
           1599: 'Hydronix Ltd',
           1600: 'AIOI- SYSTEMS CO. LTD.',
           1601: 'Ingenieurbuero Mewes & Partner GmbH',
           1602: 'HIGHYAG Lasertechnologie GmbH',
           1603: 'Ningbo Jetron Technology Co. Ltd.',
           1604: 'Myostat Motion Control',
           1605: 'A-T Controls, Inc',
           1606: 'M2M craft Co., Ltd.',
           1607: 'FUTEK Advanced Sensor Technology, Inc.',
           1608: 'Cetek',
           1609: 'Norgren Manufacturing Co., Ltd.',
           1610: 'Bernstein AG',
           1611: 'Hitachi Industrial Products, Ltd.',
           1613: 'Duplomatic MS spa',
           1614: 'Ambrit Ltd',
           1615: 'Highlight Tech Corp.',
           1616: 'New Power Plasma Co. Ltd.',
           1617: 'AGCO Corporation',
           1618: 'Techman Robot',
           1619: 'Nabeya Bi-tech Kaisha',
           1621: 'Panasonic Corporation / Electric Works Company',
           1622: 'Cytiva',
           1623: 'Janasi Industries Ltd.',
           1624: 'Haffmans BV',
           1626: 'Omniview Pty Ltd',
           1628: 'Dover Flexo Electronics',
           1630: 'IDEM Safety Switches',
           1631: 'Sonotroagel GmbH',
           1632: 'Thermo Gamma-Metrics LLC, a part of Thermo Fisher Scientific',
           1633: 'BBH Products',
           1634: 'RSI Elektrotechnik',
           1635: 'Carlo Gavazzi Ltd',
           1636: 'KOFLOC Corp.',
           1637: 'VTScada by Trihedral',
           1638: 'Hach',
           1639: 'Cogniac',
           1640: 'Toshiba Infrastructure Systems & Solutions Corporation',
           1641: 'Cannon-Automata',
           1642: 'Rosenberger',
           1643: 'Blue-White Industries',
           1644: 'Cellumation GmbH',
           1645: 'TEAC Corp.',
           1646: 'AEG Identifikationssysteme GmbH',
           1647: 'MARS TOHKEN SOLUTION CO.LTD.',
           1648: 'Midas Technology Corp.',
           1649: 'Dinkle Enterprise Co., Ltd.',
           1650: 'THALES',
           1651: 'Dunkermotoren',
           1652: 'SONOTEC GmbH',
           1653: 'Brinkmann Pumpen',
           1654: 'Rheonics',
           1655: 'Precimeter',
           1657: 'ALGO SYSTEM CO., LTD.',
           1658: 'Christ Electronic Systems GmbH',
           1659: 'JFcontrol Co., Ltd.',
           1660: 'Shenzhen Inovance Technology Co., Ltd',
           1661: 'Rheonik Coriolis Mass Flow Sensors',
           1662: 'Ichor Systems, Inc.',
           1663: 'di-soric GmbH & Co. KG',
           1664: 'Amphenol ICC',
           1665: 'Ningbo AirTAC Automation Industrial Co., Ltd.',
           1666: 'RSI',
           1667: 'Soft Robotics Inc.',
           1668: 'MUSCLE CORPORATION',
           1669: 'Spotlight Systems LLC',
           1670: 'Afag Holding AG',
           1671: 'TELESIS TECHNOLOGIES INC',
           1672: 'SSI Schaefer Automation GmbH',
           1673: 'Super Systems, Inc.',
           1674: 'CoreTigo LTD',
           1675: 'Inxpect SPA',
           1676: 'Kostal Industrie Elektrik GmbH',
           1677: 'JingQi (Tianjin) technology Co.,Ltd',
           1678: 'AGI Suretrack',
           1679: 'Shanghai JAKA Robotics Ltd',
           1680: 'Polarteknik Oy',
           1681: 'RoboteQ, Inc',
           1682: 'Uson',
           1683: 'Opt Machine Vision Tech Co., Ltd',
           1684: 'Asyril SA',
           1685: 'Georg Fischer Piping Systems',
           1686: 'Aber Instruments Ltd',
           1687: 'CodeWrights GmbH',
           1688: 'Neurala, Inc.',
           1689: 'Panasonic Software Development Center Dalian Co.,Ltd.',
           1690: 'Perinet GmbH',
           1691: 'MS Ultraschall Technologie GmbH',
           1692: 'PLASUS GmbH',
           1693: 'Nikon Corporation',
           1694: 'Shenzhen Hengzhiyuan Technology Corporation Ltd.',
           1695: 'Kowa Optronics Co., Ltd.',
           1696: 'Specialist Mechanical Engineers',
           1697: 'CMD Corporation',
           1698: 'Sanwa Engineering Corp.',
           1699: 'Intellore Systems Pvt. Ltd',
           1700: 'Toledo e Souza',
           1701: 'PBS Biotech, Inc',
           1702: 'PLANET Technology Corporation',
           1703: 'Robatech AG',
           1704: 'MARKEM-IMAJE',
           1705: 'Novanta IMS',
           1706: 'Bamboo-Dynamics',
           1707: 'FACTS Engineering, LLC',
           1708: 'Digital Dynamics',
           1709: 'Fatek Automation Corporation',
           1710: 'Hanwha',
           1711: 'Fukuda',
           1712: 'Zhejiang Eternal Automation Sci-Tec Co.,Ltd',
           1713: 'KYOWA ELECTRONIC INSTRUMENTS CO.,LTD.',
           1714: 'Vaisala Oyj',
           1715: 'Hennecke GmbH',
           1716: 'Encoder Products Company',
           1717: 'Converting Equipment International (dba CEI)',
           1719: 'Kinova',
           1720: 'The Poling Group, Inc.',
           1721: 'plating electronic GmbH',
           1722: 'HIWIN MIKROSYSTEM CORP.',
           1723: 'Wuxi Xinje Electric Co.,Ltd.',
           1724: 'ViSCO Technologies Corporation',
           1725: 'MinebeaMitsumi Inc.',
           1726: 'FIAtec GmbH',
           1727: 'eSOL Co.,Ltd',
           1728: 'NTN TECHNICAL SERVICE',
           1729: 'Shanghai Flexem',
           1730: 'Magswitch',
           1731: 'VEGA Grieshaber KG',
           1732: 'H.D.T. S.R.L.',
           1733: 'Tool-Temp AG',
           1734: 'Hollysys Technology Group Co., Ltd',
           9876: 'ODVA'}
